import os
os.environ['CUDA_VISIBLE_DEVICES'] = '3'

import numpy as np
import torch
import torchvision.transforms as T
from PIL import Image
from torchvision.transforms.functional import InterpolationMode
from transformers import AutoModel, AutoTokenizer
import requests
from io import BytesIO
from urllib.parse import urlparse
from torch.profiler import profile, record_function, ProfilerActivity
import time
import pickle
import random
from datetime import datetime

IMAGENET_MEAN = (0.485, 0.456, 0.406)
IMAGENET_STD = (0.229, 0.224, 0.225)

def build_transform(input_size):
    MEAN, STD = IMAGENET_MEAN, IMAGENET_STD
    transform = T.Compose([
        T.Lambda(lambda img: img.convert('RGB') if img.mode != 'RGB' else img),
        T.Resize((input_size, input_size), interpolation=InterpolationMode.BICUBIC),
        T.ToTensor(),
        T.Normalize(mean=MEAN, std=STD)
    ])
    return transform

def find_closest_aspect_ratio(aspect_ratio, target_ratios, width, height, image_size):
    best_ratio_diff = float('inf')
    best_ratio = (1, 1)
    area = width * height
    for ratio in target_ratios:
        target_aspect_ratio = ratio[0] / ratio[1]
        ratio_diff = abs(aspect_ratio - target_aspect_ratio)
        if ratio_diff < best_ratio_diff:
            best_ratio_diff = ratio_diff
            best_ratio = ratio
        elif ratio_diff == best_ratio_diff:
            if area > 0.5 * image_size * image_size * ratio[0] * ratio[1]:
                best_ratio = ratio
    return best_ratio

def dynamic_preprocess(image, min_num=1, max_num=12, image_size=448, use_thumbnail=False):
    orig_width, orig_height = image.size
    aspect_ratio = orig_width / orig_height

    # calculate the existing image aspect ratio
    target_ratios = set(
        (i, j) for n in range(min_num, max_num + 1) for i in range(1, n + 1) for j in range(1, n + 1) if
        i * j <= max_num and i * j >= min_num)
    target_ratios = sorted(target_ratios, key=lambda x: x[0] * x[1])

    # find the closest aspect ratio to the target
    target_aspect_ratio = find_closest_aspect_ratio(
        aspect_ratio, target_ratios, orig_width, orig_height, image_size)

    # calculate the target width and height
    target_width = image_size * target_aspect_ratio[0]
    target_height = image_size * target_aspect_ratio[1]
    blocks = target_aspect_ratio[0] * target_aspect_ratio[1]

    # resize the image
    resized_img = image.resize((target_width, target_height))
    processed_images = []
    for i in range(blocks):
        box = (
            (i % (target_width // image_size)) * image_size,
            (i // (target_width // image_size)) * image_size,
            ((i % (target_width // image_size)) + 1) * image_size,
            ((i // (target_width // image_size)) + 1) * image_size
        )
        # split the image
        split_img = resized_img.crop(box)
        processed_images.append(split_img)
    assert len(processed_images) == blocks
    if use_thumbnail and len(processed_images) != 1:
        thumbnail_img = image.resize((image_size, image_size))
        processed_images.append(thumbnail_img)
    return processed_images

def load_image(image_file, input_size=448, max_num=12):
    # 如果已经是 PIL Image 对象
    if isinstance(image_file, Image.Image):
        image = image_file.convert('RGB')
    # 如果是 URL
    elif isinstance(image_file, str) and bool(urlparse(image_file).netloc):
        try:
            response = requests.get(image_file, timeout=10)
            response.raise_for_status()  # 检查请求是否成功
            image = Image.open(BytesIO(response.content)).convert('RGB')
        except Exception as e:
            raise ValueError(f"无法从URL加载图片: {str(e)}")
    # 如果是本地文件路径
    elif isinstance(image_file, str):
        image = Image.open(image_file).convert('RGB')
    # 如果是字节数据
    elif isinstance(image_file, bytes):
        image = Image.open(BytesIO(image_file)).convert('RGB')
    else:
        raise ValueError(f"不支持的图片格式: {type(image_file)}")

    transform = build_transform(input_size=input_size)
    images = dynamic_preprocess(image, image_size=input_size, use_thumbnail=True, max_num=max_num)
    pixel_values = [transform(image) for image in images]
    pixel_values = torch.stack(pixel_values)
    return pixel_values

# If you want to load a model using multiple GPUs, please refer to the `Multiple GPUs` section.
path = 'Internvl2_5'
model = AutoModel.from_pretrained(
    path,
    torch_dtype=torch.bfloat16,
    low_cpu_mem_usage=True,
    use_flash_attn=True,
    trust_remote_code=True,
    vision_model = None,
    language_model = None).eval().cuda()

# 加载要测试的权重
state_dict = torch.load("mlp_epoch_5.pth")  # 加载训练好的权重
model.load_state_dict(state_dict)
generation_config = dict(max_new_tokens=1024, do_sample=True)
tokenizer = AutoTokenizer.from_pretrained(path, trust_remote_code=True, use_fast=False)

print(model)

# 测试
# 加载数据
with open('batch_1.pkl', 'rb') as f:
    data = pickle.load(f)

# 随机选择100条数据
test_samples = random.sample(data, 100)

# 创建结果文件
timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
results_file = f'test_results_{timestamp}.txt'

with open(results_file, 'w', encoding='utf-8') as f:
    for i, test_item in enumerate(test_samples, 1):
        question = "Identify the brand of the product in the picture, and write a caption including brand information in 200 words."
        expected_answer = test_item['response']
        
        try:
            pixel_values = load_image(test_item['image'], max_num=12).to(torch.bfloat16).cuda()
            response, _ = model.chat(tokenizer, pixel_values, question, generation_config, history=None, return_history=True)
            
            # 写入结果
            f.write(f"Test case {i}/100:\n")
            f.write(f"Question: {question}\n")
            f.write(f"Expected: {expected_answer}\n")
            f.write(f"Response: {response}\n")
            f.write("-" * 50 + "\n")
            
            # 打印进度
            print(f"Processed {i}/100 samples")
            
        except Exception as e:
            f.write(f"Test case {i}/100 - ERROR:\n")
            f.write(f"Error message: {str(e)}\n")
            f.write("-" * 50 + "\n")
            print(f"Error processing sample {i}: {str(e)}")

print(f"测试完成！结果已保存到文件：{results_file}")
